@@ -7,6 +7,7 @@ from sqlalchemy import *
 
 from migrate import changeset, exceptions
 from migrate.changeset import *
+from migrate.changeset import constraint
 from migrate.changeset.schema import ColumnDelta
 from migrate.tests import fixture
 from migrate.tests.fixture.warnings import catch_warnings
@@ -17,6 +18,7 @@ class TestAddDropColumn(fixture.DB):
     """
     level = fixture.DB.CONNECT
     table_name = 'tmp_adddropcol'
+    table_name_idx = 'tmp_adddropcol_idx'
     table_int = 0
 
     def _setup(self, url):
@@ -25,14 +27,27 @@ class TestAddDropColumn(fixture.DB):
         self.table = Table(self.table_name, self.meta,
             Column('id', Integer, unique=True),
         )
+        self.table_idx = Table(
+            self.table_name_idx,
+            self.meta,
+            Column('id', Integer, primary_key=True),
+            Column('a', Integer),
+            Column('b', Integer),
+            Index('test_idx', 'a', 'b')
+        )
         self.meta.bind = self.engine
         if self.engine.has_table(self.table.name):
             self.table.drop()
+        if self.engine.has_table(self.table_idx.name):
+            self.table_idx.drop()
         self.table.create()
+        self.table_idx.create()
 
     def _teardown(self):
         if self.engine.has_table(self.table.name):
             self.table.drop()
+        if self.engine.has_table(self.table_idx.name):
+            self.table_idx.drop()
         self.meta.clear()
         super(TestAddDropColumn,self)._teardown()
 
@@ -44,11 +59,11 @@ class TestAddDropColumn(fixture.DB):
             self.refresh_table(self.table_name)
             result = len(self.table.c)
 
-            self.assertEquals(result, num_of_expected_cols),
+            self.assertEqual(result, num_of_expected_cols),
             if col_k.get('primary_key', None):
                 # new primary key: check its length too
                 result = len(self.table.primary_key)
-                self.assertEquals(result, num_of_expected_cols)
+                self.assertEqual(result, num_of_expected_cols)
 
         # we have 1 columns and there is no data column  
         assert_numcols(1)
@@ -168,9 +183,7 @@ class TestAddDropColumn(fixture.DB):
         reftable.create()
 
         # create column with fk
-        col = Column('data', Integer, ForeignKey(reftable.c.id))
-        if SQLA_07:
-            self.table.append_column(col)
+        col = Column('data', Integer, ForeignKey(reftable.c.id, name='testfk'))
         col.create(self.table)
 
         # check if constraint is added
@@ -188,6 +201,11 @@ class TestAddDropColumn(fixture.DB):
         else:
             self.assertEqual(reftable.c.id.name,
                 col.foreign_keys[0].column.name)
+
+        if self.engine.name == 'mysql':
+            constraint.ForeignKeyConstraint([self.table.c.data],
+                                            [reftable.c.id],
+                                            name='testfk').drop()
         col.drop(self.table)
 
         if self.engine.has_table(reftable.name):
@@ -213,18 +231,18 @@ class TestAddDropColumn(fixture.DB):
 
         col.drop()
 
-    @fixture.usedb(not_supported='mysql')
+    @fixture.usedb(not_supported=['mysql'])
     def test_check(self):
         """Can create columns with check constraint"""
-        col = Column('data',
+        col = Column('foo',
                      Integer,
-                     sqlalchemy.schema.CheckConstraint('data > 4'))
+                     sqlalchemy.schema.CheckConstraint('foo > 4'))
         col.create(self.table)
 
         # check if constraint was added (cannot test on objects)
-        self.table.insert(values={'data': 5}).execute()
+        self.table.insert(values={'foo': 5}).execute()
         try:
-            self.table.insert(values={'data': 3}).execute()
+            self.table.insert(values={'foo': 3}).execute()
         except (sqlalchemy.exc.IntegrityError,
                 sqlalchemy.exc.ProgrammingError):
             pass
@@ -253,7 +271,27 @@ class TestAddDropColumn(fixture.DB):
 
         col.drop(self.table)
 
-# TODO: remove already attached columns with indexes, uniques, pks, fks ..
+# TODO: remove already attached columns with uniques, pks, fks ..
+    @fixture.usedb(not_supported=['ibm_db_sa', 'postgresql'])
+    def test_drop_column_of_composite_index(self):
+        # NOTE(rpodolyaka): postgresql automatically drops a composite index
+        #                   if one of its columns is dropped
+        # NOTE(mriedem): DB2 does the same.
+        self.table_idx.c.b.drop()
+
+        reflected = Table(self.table_idx.name, MetaData(), autoload=True,
+                          autoload_with=self.engine)
+        index = next(iter(reflected.indexes))
+        self.assertEquals(['a'], [c.name for c in index.columns])
+
+    @fixture.usedb()
+    def test_drop_all_columns_of_composite_index(self):
+        self.table_idx.c.a.drop()
+        self.table_idx.c.b.drop()
+
+        reflected = Table(self.table_idx.name, MetaData(), autoload=True,
+                          autoload_with=self.engine)
+        self.assertEquals(0, len(reflected.indexes))
 
     def _check_index(self,expected):
         if 'mysql' in self.engine.name or 'postgres' in self.engine.name:
@@ -339,26 +377,21 @@ class TestAddDropColumn(fixture.DB):
         self.table.create()
 
         # paranoid check
-        if SQLA_06:
-            self.refresh_table()
-            self.assertEqual(
-                sorted([i.name for i in self.table.indexes]),
-                [u'ix_tmp_adddropcol_d1', u'ix_tmp_adddropcol_d2']
-                )
+        self.refresh_table()
+        self.assertEqual(
+            sorted([i.name for i in self.table.indexes]),
+            [u'ix_tmp_adddropcol_d1', u'ix_tmp_adddropcol_d2']
+            )
             
         # delete one
         self.table.c.d2.drop()
 
         # ensure the other index is still there
-        if SQLA_06:
-            self.refresh_table()
-            self.assertEqual(
-                sorted([i.name for i in self.table.indexes]),
-                [u'ix_tmp_adddropcol_d1']
-                )
-        else:
-            # a crude test for 0.5.x
-            Index('ix_tmp_adddropcol_d1',self.table.c.d1).drop()
+        self.refresh_table()
+        self.assertEqual(
+            sorted([i.name for i in self.table.indexes]),
+            [u'ix_tmp_adddropcol_d1']
+            )
             
     def _actual_foreign_keys(self):
         from sqlalchemy.schema import ForeignKeyConstraint
@@ -391,8 +424,8 @@ class TestAddDropColumn(fixture.DB):
         self.table = Table(
             self.table_name, self.meta,
             Column('id', Integer, primary_key=True),
-            Column('r1', Integer, ForeignKey('tmp_ref.id')),
-            Column('r2', Integer, ForeignKey('tmp_ref.id')),
+            Column('r1', Integer, ForeignKey('tmp_ref.id', name='test_fk1')),
+            Column('r2', Integer, ForeignKey('tmp_ref.id', name='test_fk2')),
             )
         self.table.create()
 
@@ -401,12 +434,15 @@ class TestAddDropColumn(fixture.DB):
                          self._actual_foreign_keys())
         
         # delete one
+        if self.engine.name == 'mysql':
+            constraint.ForeignKeyConstraint([self.table.c.r2], [reftable.c.id],
+                                            name='test_fk2').drop()
         self.table.c.r2.drop()
         
         # check remaining foreign key is there
         self.assertEqual([['r1']],
                          self._actual_foreign_keys())
-        
+
     @fixture.usedb()
     def test_drop_with_complex_foreign_keys(self):
         from sqlalchemy.schema import ForeignKeyConstraint
@@ -414,11 +450,16 @@ class TestAddDropColumn(fixture.DB):
         
         self.table.drop()
         self.meta.clear()
-        
+
+        # NOTE(mriedem): DB2 does not currently support unique constraints
+        # on nullable columns, so the columns that are used to create the
+        # foreign keys here need to be non-nullable for testing with DB2
+        # to work.
+
         # create FK's target
         reftable = Table('tmp_ref', self.meta,
             Column('id', Integer, primary_key=True),
-            Column('jd', Integer),
+            Column('jd', Integer, nullable=False),
             UniqueConstraint('id','jd')
             )
         if self.engine.has_table(reftable.name):
@@ -429,8 +470,8 @@ class TestAddDropColumn(fixture.DB):
         self.table = Table(
             self.table_name, self.meta,
             Column('id', Integer, primary_key=True),
-            Column('r1', Integer),
-            Column('r2', Integer),
+            Column('r1', Integer, nullable=False),
+            Column('r2', Integer, nullable=False),
             ForeignKeyConstraint(['r1','r2'],
                                  [reftable.c.id,reftable.c.jd],
                                  name='test_fk')
@@ -442,6 +483,10 @@ class TestAddDropColumn(fixture.DB):
                          self._actual_foreign_keys())
         
         # delete one
+        if self.engine.name == 'mysql':
+            constraint.ForeignKeyConstraint([self.table.c.r1, self.table.c.r2],
+                                            [reftable.c.id, reftable.c.jd],
+                                            name='test_fk').drop()
         self.table.c.r2.drop()
         
         # check the constraint is gone, since part of it
@@ -492,7 +537,7 @@ class TestRename(fixture.DB):
             """
             if not skip_object_check:
                 # Table object check
-                self.assertEquals(self.table.name,expected)
+                self.assertEqual(self.table.name,expected)
                 newname = self.table.name
             else:
                 # we know the object's name isn't consistent: just assign it
@@ -500,20 +545,28 @@ class TestRename(fixture.DB):
             # Table DB check
             self.meta.clear()
             self.table = Table(newname, self.meta, autoload=True)
-            self.assertEquals(self.table.name, expected)
+            self.assertEqual(self.table.name, expected)
 
         def assert_index_name(expected, skip_object_check=False):
             if not skip_object_check:
                 # Index object check
-                self.assertEquals(self.index.name, expected)
+                self.assertEqual(self.index.name, expected)
             else:
                 # object is inconsistent
                 self.index.name = expected
             # TODO: Index DB check
-            
+
+        def add_table_to_meta(name):
+            # trigger the case where table_name2 needs to be
+            # removed from the metadata in ChangesetTable.deregister()
+            tmp = Table(name, self.meta, Column(c_name, Integer))
+            tmp.create()
+            tmp.drop()
+
         try:
             # Table renames
             assert_table_name(table_name1)
+            add_table_to_meta(table_name2)
             rename_table(self.table, table_name2)
             assert_table_name(table_name2)
             self.table.rename(table_name1)
@@ -559,7 +612,7 @@ class TestColumnChange(fixture.DB):
             self.table.drop()
         try:
             self.table.create()
-        except sqlalchemy.exceptions.SQLError, e:
+        except sqlalchemy.exc.SQLError, e:
             # SQLite: database schema has changed
             if not self.url.startswith('sqlite://'):
                 raise
@@ -568,7 +621,7 @@ class TestColumnChange(fixture.DB):
         if self.table.exists():
             try:
                 self.table.drop(self.engine)
-            except sqlalchemy.exceptions.SQLError,e:
+            except sqlalchemy.exc.SQLError,e:
                 # SQLite: database schema has changed
                 if not self.url.startswith('sqlite://'):
                     raise
@@ -582,7 +635,7 @@ class TestColumnChange(fixture.DB):
         # Table content should be preserved in changed columns
         content = "fgsfds"
         self.engine.execute(self.table.insert(), data=content, id=42)
-        self.assertEquals(num_rows(self.table.c.data, content), 1)
+        self.assertEqual(num_rows(self.table.c.data, content), 1)
 
         # ...as a function, given a column object and the new name
         alter_column('data', name='data2', table=self.table)
@@ -591,14 +644,14 @@ class TestColumnChange(fixture.DB):
         self.refresh_table(self.table.name)
         self.assert_('data' not in self.table.c.keys())
         self.assert_('atad' in self.table.c.keys())
-        self.assertEquals(num_rows(self.table.c.atad, content), 1)
+        self.assertEqual(num_rows(self.table.c.atad, content), 1)
 
         # ...as a method, given a new name
         self.table.c.atad.alter(name='data')
         self.refresh_table(self.table.name)
         self.assert_('atad' not in self.table.c.keys())
         self.table.c.data # Should not raise exception
-        self.assertEquals(num_rows(self.table.c.data, content), 1)
+        self.assertEqual(num_rows(self.table.c.data, content), 1)
 
         # ...as a function, given a new object
         alter_column(self.table.c.data,
@@ -607,7 +660,7 @@ class TestColumnChange(fixture.DB):
         self.refresh_table(self.table.name)
         self.assert_('data' not in self.table.c.keys())
         self.table.c.atad   # Should not raise exception
-        self.assertEquals(num_rows(self.table.c.atad, content), 1)
+        self.assertEqual(num_rows(self.table.c.atad, content), 1)
 
         # ...as a method, given a new object
         self.table.c.atad.alter(
@@ -617,7 +670,7 @@ class TestColumnChange(fixture.DB):
         self.refresh_table(self.table.name)
         self.assert_('atad' not in self.table.c.keys())
         self.table.c.data   # Should not raise exception
-        self.assertEquals(num_rows(self.table.c.data,content), 1)
+        self.assertEqual(num_rows(self.table.c.data,content), 1)
         
     @fixture.usedb()
     def test_type(self):
@@ -627,15 +680,15 @@ class TestColumnChange(fixture.DB):
         self.table.c.data.alter(type=String(43))
         self.refresh_table(self.table.name)
         self.assert_(isinstance(self.table.c.data.type, String))
-        self.assertEquals(self.table.c.data.type.length, 43)
+        self.assertEqual(self.table.c.data.type.length, 43)
 
         # Different type
         self.assert_(isinstance(self.table.c.id.type, Integer))
-        self.assertEquals(self.table.c.id.nullable, False)
+        self.assertEqual(self.table.c.id.nullable, False)
 
         if not self.engine.name == 'firebird':
             self.table.c.id.alter(type=String(20))
-            self.assertEquals(self.table.c.id.nullable, False)
+            self.assertEqual(self.table.c.id.nullable, False)
             self.refresh_table(self.table.name)
             self.assert_(isinstance(self.table.c.id.type, String))
 
@@ -645,13 +698,13 @@ class TestColumnChange(fixture.DB):
         Only DefaultClauses are changed here: others are managed by the 
         application / by SA
         """
-        self.assertEquals(self.table.c.data.server_default.arg, 'tluafed')
+        self.assertEqual(self.table.c.data.server_default.arg, 'tluafed')
 
         # Just the new default 
         default = 'my_default'
         self.table.c.data.alter(server_default=DefaultClause(default))
         self.refresh_table(self.table.name)
-        #self.assertEquals(self.table.c.data.server_default.arg,default)
+        #self.assertEqual(self.table.c.data.server_default.arg,default)
         # TextClause returned by autoload
         self.assert_(default in str(self.table.c.data.server_default.arg))
         self.engine.execute(self.table.insert(), id=12)
@@ -672,27 +725,24 @@ class TestColumnChange(fixture.DB):
         # server_default isn't necessarily None for Oracle
         #self.assert_(self.table.c.data.server_default is None,self.table.c.data.server_default)
         self.engine.execute(self.table.insert(), id=11)
-        if SQLA_06:
-            row = self.table.select(self.table.c.id == 11).execution_options(autocommit=True).execute().fetchone()
-        else:
-            row = self.table.select(self.table.c.id == 11, autocommit=True).execute().fetchone()
+        row = self.table.select(self.table.c.id == 11).execution_options(autocommit=True).execute().fetchone()
         self.assert_(row['data'] is None, row['data'])
 
     @fixture.usedb(not_supported='firebird')
     def test_null(self):
         """Can change a column's null constraint"""
-        self.assertEquals(self.table.c.data.nullable, True)
+        self.assertEqual(self.table.c.data.nullable, True)
 
         # Full column
         self.table.c.data.alter(type=String(40), nullable=False)
         self.table.nullable = None
         self.refresh_table(self.table.name)
-        self.assertEquals(self.table.c.data.nullable, False)
+        self.assertEqual(self.table.c.data.nullable, False)
 
         # Just the new status
         self.table.c.data.alter(nullable=True)
         self.refresh_table(self.table.name)
-        self.assertEquals(self.table.c.data.nullable, True)
+        self.assertEqual(self.table.c.data.nullable, True)
 
     @fixture.usedb()
     def test_alter_deprecated(self):
@@ -795,7 +845,7 @@ class TestColumnDelta(fixture.DB):
         self.delta = ColumnDelta(original, *p, **k)
         result = self.delta.keys()
         result.sort()
-        self.assertEquals(expected, result)
+        self.assertEqual(expected, result)
         return self.delta
 
     def test_deltas_two_columns(self):
@@ -887,8 +937,8 @@ class TestColumnDelta(fixture.DB):
 
         # Change name, given an up-to-date definition and the current name
         delta = self.verify(['name'], col_orig, name='blah')
-        self.assertEquals(delta.get('name'), 'blah')
-        self.assertEquals(delta.current_name, 'id')
+        self.assertEqual(delta.get('name'), 'blah')
+        self.assertEqual(delta.current_name, 'id')
 
         col_orig = self.mkcol(primary_key=True)
         self.verify(['name', 'type'], col_orig, name='id12', type=Text, alter_metadata=True)
